"""Holds data models for the whoop api.

Copyright (c) 2022 Felix Geilert
"""


from abc import abstractclassmethod
from datetime import datetime
from typing import Dict
from pydantic import BaseModel
import time_helper as th


class UserProfile(BaseModel):
    """Represents a user profile."""

    user_id: int
    email: str
    first_name: str
    last_name: str


class UserMeasurements(BaseModel):
    """Stores user measurements."""

    height_meter: float
    weight_kilogram: float
    max_heart_rate: int


class UserData(BaseModel):
    """Base Data that is used for most user data objects"""

    user_id: int
    created_at: datetime
    updated_at: datetime
    timezone_offset: str = None
    score_state: str

    @abstractclassmethod
    def _dict_parse(cls, data: Dict):
        return data

    @classmethod
    def from_dict(cls, data: Dict):
        for dt in ["created_at", "updated_at", "start", "end"]:
            if dt in data:
                data[dt] = th.any_to_datetime(data[dt])
        data = cls._dict_parse(data)

        return cls(**data)


class UserCycleScore(BaseModel):
    strain: float
    kilojoule: float
    average_heart_rate: int
    max_heart_rate: int

    @property
    def calories(self):
        return self.kilojoule / 4.184


class UserCycle(UserData):
    """Stores data of the user cycle."""

    id: int
    start: datetime
    end: datetime = None
    score: UserCycleScore = None

    @classmethod
    def _dict_parse(cls, data: Dict):
        if "score" in data and data["score"] is not None:
            data["score"] = UserCycleScore(**data["score"])

        return data


class UserSleepStages(BaseModel):
    """Stores the sleep stages."""

    total_in_bed_time_milli: int
    total_awake_time_milli: int
    total_no_data_time_milli: int
    total_light_sleep_time_milli: int
    total_slow_wave_sleep_time_milli: int
    total_rem_sleep_time_milli: int
    sleep_cycle_count: int
    disturbance_count: int


class UserSleepNeed(BaseModel):
    """Stores the sleep need."""

    baseline_milli: int
    need_from_sleep_debt_milli: int
    need_from_recent_strain_milli: int
    need_from_recent_nap_milli: int


class UserSleepScore(BaseModel):
    """Stores the score of the user sleep."""

    stage_summary: UserSleepStages = None
    sleep_needed: UserSleepNeed = None
    respiratory_rate: float = None
    sleep_performance_percentage: float = None
    sleep_consistency_percentage: float = None
    sleep_efficiency_percentage: float = None


class UserSleep(UserData):
    """Stores data for user sleep."""

    id: int
    nap: bool
    start: datetime
    end: datetime = None
    score: UserSleepScore = None

    @classmethod
    def _dict_parse(cls, data: Dict):
        if "score" in data and data["score"] is not None:
            score_dict = data["score"]
            if (
                "stage_summary" in score_dict
                and score_dict["stage_summary"] is not None
            ):
                score_dict["stage_summary"] = UserSleepStages(
                    **score_dict["stage_summary"]
                )
            else:
                score_dict["stage_summary"] = None
            if "sleep_needed" in score_dict and score_dict["sleep_needed"] is not None:
                score_dict["sleep_needed"] = UserSleepNeed(**score_dict["sleep_needed"])
            else:
                score_dict["sleep_needed"] = None
            data["score"] = UserSleepScore(**score_dict)

        return data


class UserRecoveryScore(BaseModel):
    """Score of users recovery."""

    user_calibrating: bool
    recovery_score: float
    resting_heart_rate: float
    hrv_rmssd_milli: float
    spo2_percentage: float = None
    skin_temp_celsius: float = None


class UserRecovery(UserData):
    """Stores data for user recovery."""

    cycle_id: int
    sleep_id: int
    score: UserRecoveryScore

    @classmethod
    def _dict_parse(cls, data: Dict):
        if "score" in data and data["score"] is not None:
            data["score"] = UserRecoveryScore(**data["score"])

        return data


class UserWorkoutZoneDuration(BaseModel):
    """Stores the duration of the workout zone."""

    zone_zero_milli: int
    zone_one_milli: int
    zone_two_milli: int
    zone_three_milli: int
    zone_four_milli: int
    zone_five_milli: int


class UserWorkoutScore(BaseModel):
    """Score of the user workout."""

    strain: float
    average_heart_rate: int
    max_heart_rate: int
    kilojoule: float
    percent_recorded: float
    distance_meter: float = None
    altitude_gain_meter: float = None
    altitude_change_meter: float = None
    zone_duration: UserWorkoutZoneDuration = None


class UserWorkout(UserData):
    """Stores information about a user workout"""

    id: int
    start: datetime
    end: datetime = None
    sport_id: int
    score: UserWorkoutScore = None

    @classmethod
    def _dict_parse(cls, data: Dict):
        if "score" in data and data["score"] is not None:
            score_dict = data["score"]
            if (
                "zone_duration" in score_dict
                and score_dict["zone_duration"] is not None
            ):
                score_dict["zone_duration"] = UserWorkoutZoneDuration(
                    **score_dict["zone_duration"]
                )
            data["score"] = UserWorkoutScore(**score_dict)

        return data


SPORT_IDS = {
    -1: "Activity",
    126: "Assault Bike",
    85: "Australian Football",
    107: "Barre",
    16: "Baseball",
    17: "Basketball",
    10: "Box Fitness",
    39: "Boxing",
    93: "Caddying",
    46: "Canoeing",
    113: "Circus Arts",
    83: "Climber",
    87: "Coaching",
    89: "Commuting",
    100: "Cricket",
    47: "Cross Country Skiing",
    1: "Cycling",
    42: "Dance",
    73: "Diving",
    49: "Duathlon",
    65: "Elliptical",
    19: "Fencing",
    20: "Field Hockey",
    21: "Football",
    48: "Functional Fitness",
    111: "Gaelic Football",
    90: "Gaming",
    22: "Golf",
    51: "Gymnastics",
    96: "HIIT",
    109: "High Stress Work",
    52: "Hiking/Rucking",
    53: "Horseback Riding",
    112: "Hurling/Camogie",
    88: "Ice Bath",
    24: "Ice Hockey",
    102: "Inline Skating",
    98: "Jiu Jitsu",
    54: "Jogging",
    84: "Jumping Rope",
    55: "Kayaking",
    127: "Kickboxing",
    25: "Lacrosse",
    50: "Machine Workout",
    99: "Manual Labor",
    56: "Martial Arts",
    121: "Massage Therapy",
    70: "Meditation",
    92: "Motocross",
    95: "Motor Racing",
    57: "Mountain Biking",
    94: "Obstacle Course Racing",
    58: "Obstacle Racing",
    76: "Operations - Flying",
    75: "Operations - Medical",
    74: "Operations - Tactical",
    77: "Operations - Water",
    71: "Other",
    131: "Other - Recovery",
    106: "Paddle Tennis",
    61: "Paddleboarding",
    110: "Parkour",
    101: "Pickleball",
    43: "Pilates",
    72: "Pit Practice",
    67: "Plyometrics",
    59: "Powerlifting",
    116: "Resonance Frequency Breathing",
    60: "Rock Climbing",
    18: "Rowing",
    27: "Rugby",
    0: "Running",
    28: "Sailing",
    69: "Sex",
    86: "Skateboarding",
    29: "Skiing",
    91: "Snowboarding",
    30: "Soccer",
    31: "Softball",
    104: "Spikeball",
    97: "Spin",
    68: "Spinning",
    32: "Squash",
    108: "Stage Performance",
    66: "Stairmaster",
    128: "Stretching",
    64: "Surfing",
    33: "Swimming",
    34: "Tennis",
    35: "Track & Field",
    62: "Triathlon",
    82: "Ultimate",
    36: "Volleyball",
    63: "Walking",
    125: "Watching Sports",
    37: "Water Polo",
    45: "Weightlifting",
    105: "Wheelchair Pushing",
    38: "Wrestling",
    44: "Yoga",
}



__version__ = "0.1.2"

import logging

try:
    # from . import auth
    from .client_v1 import WhoopClient, API_VERSION
    from .models.models_v1 import SPORT_IDS

    # import versions
    from . import client_vu7
    from . import client_v1
except Exception as ex:
    logging.error(f"Not all dependencies installed: {ex}")



"""Main Client for retrieving Whoop data

This contains information for the unofficial version 7 of the Whoop API.

Copyright (C) 2022 Felix Geilert
"""
import json
from typing import Dict, Union

import requests
import configparser
import logging
from datetime import datetime

from dateutil import parser
import pandas as pd
import numpy as np
from time_helper import create_intervals, localize_datetime


# define default whoop date format
DATE_FORMAT = "%Y-%m-%d"

# define API specifications
API_VERSION = "7"
API_URL = f"https://api-{API_VERSION}.whoop.com"

# columns that should be converted
STR_COLS = ["strain.workouts"]


class AuthenticationError(Exception):
    pass


def whoop_time_str(dt: Union[datetime, str]):
    """Converts a datetime object into a whoop timestring"""
    # convert to utc
    dt_utc = localize_datetime(dt, "Etc/UTC")
    ms = int(str(dt_utc.microsecond)[0:3])
    return f"{dt_utc.strftime(DATE_FORMAT)}T{dt_utc.strftime('%H:%M:%S')}.{ms:03}Z"


class WhoopClient:
    """
    A class object to allow a user to login and store their authorization code,
    then perform pulls using the code in order to access different types of data

    Args:
        auth_code (str): Authorization Code for whoop login
        whoop_id (str): Username
    """

    def __init__(
        self,
        auth_token=None,
        whoop_id=None,
        refresh_token=None,
        current_datetime=datetime.utcnow(),
    ):
        # create some general params
        self.auth_token = auth_token
        self.refresh_token = refresh_token
        self.whoop_id = whoop_id
        self.current_datetime = current_datetime
        self.start_datetime = None
        self.all_activities = None
        self.sport_dict = None
        self.all_sleep_events = None

        # check if whoop id should be pulled
        if self.auth_token and not self.start_datetime:
            self.pull_userinfo()

    def _create_url(self, postfix="", auth=False, user=True):
        """Generates the URL of the whoop endpoints"""
        base_url = API_URL

        # check for auth
        if auth:
            return f"{base_url}/oauth/token"

        # check for user
        if user:
            if not self.whoop_id:
                raise ValueError("No whoop user has been authenticated!")
            base_url = f"{API_URL}/users/{self.whoop_id}"

        # generate url
        return f"{base_url}/{postfix}"

    def pull_api(self, url, params=None, df=False):
        """Generalized function to retrieve data from the API.

        Args:
            url: Path that should be pulled
            params: Query Parameters to be passed
            df: Defines if the output data should be parsed as dataframe
        """
        # provides the authorization code for the header (if provided)
        headers = {}
        if self.auth_token:
            headers["authorization"] = f"Bearer {self.auth_token}"

        # send the request
        pull = requests.get(url, params=params, headers=headers)

        # retrieve json data from the API
        if pull.status_code == 200 and len(pull.content) > 1:
            if df:
                d = pd.json_normalize(pull.json())

                # convert potential str columns
                for col in STR_COLS:
                    # make sure column exists
                    if col not in d.columns:
                        continue

                    # check for type
                    if d[col].dtype == "object":
                        d[col] = d[col].apply(
                            lambda x: json.loads(
                                str(x).replace("'", '"').replace('u"', '"')
                            )
                        )

                return d
            else:
                return pull.json()
        elif pull.status_code == 401 and self.auth_refresh < 3:
            raise AuthenticationError("Unable to authenticate wiht backend")
        else:
            raise IOError(
                f"Unable to retrieve API response for url ({url}): {pull.status_code}",
                pull,
                params,
            )

    def _parse_authentication(self, body):
        """Parses the authentication response"""
        self.whoop_id = body["user"]["id"]
        self.auth_token = body["access_token"]
        self.refresh_token = body.get("refresh_token")
        start_time = body["user"]["profile"]["createdAt"]
        self.start_datetime = parser.isoparse(start_time)
        logging.info("Authentication successful")

        return self.get_auth()

    def authenticate_ini(self, ini_file):
        # read the user data provided in ini
        config = configparser.ConfigParser()
        config.read(ini_file)
        username = config["whoop"]["username"]
        password = config["whoop"]["password"]

        # pass along
        return self.authenticate_user(username, password)

    def authenticate_user(self, username, password):
        # create header for token request
        headers = {
            "username": username,
            "password": password,
            "grant_type": "password",
            "issueRefresh": False,
        }

        # send request
        auth = requests.post(self._create_url(auth=True), json=headers)

        # check for errors
        if auth.status_code != 200:
            raise AuthenticationError(f"Unable to complete authentication: {auth.text}")

        # parse the data
        return self._parse_authentication(auth.json())

    def authenticate_refresh(self, refresh_token=None):
        # retrieve token
        refresh_token = refresh_token if refresh_token else self.refresh_token

        # check if token is set
        if not refresh_token:
            raise ValueError("No Refresh Token provided")

        headers = {
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
            "issueRefresh": True,
        }

        # send request
        auth = self.pull_api(self._create_url(auth=True), params=headers)

        # check for errors
        if auth.status_code != 200:
            raise AuthenticationError(f"Unable to complete authentication: {auth.text}")

        # parse the data
        return self._parse_authentication(auth.json())

    def authenticate_code(self, auth_code):
        raise NotImplementedError("Public APIs currently unsupported by whoop")

    def get_auth(self):
        """Returns current authentication data"""
        return (
            self.auth_token,
            self.refresh_token,
            {"user_id": self.whoop_id, "createdAt": self.start_datetime},
        )

    def pull_userinfo(self):
        """Retrieves user information based on the whoop_id"""
        data = self.pull_api(self._create_url())
        start_time = data["createdAt"]
        self.start_datetime = parser.isoparse(start_time)
        return data

    def pull_sleep_main(self, sleep_id):
        sleep = self.pull_api(self._create_url(f"sleeps/{sleep_id}"))

        # retrieve the data
        main_df = pd.json_normalize(sleep)
        return main_df

    def pull_sleep_events(self, sleep_id):
        sleep = self.pull_api(self._create_url(f"sleeps/{sleep_id}"))

        # retrieve the data
        events_df = pd.json_normalize(sleep["events"])
        events_df["id"] = sleep_id
        return events_df

    def get_keydata_raw(self, start=None, end=None):
        """Retrieves all data as array of raw jsons from the creation time of the user."""
        # store all results
        results = []

        # starts crawling from user creation-time
        if self.start_datetime:
            # parse the data
            if start is not None:
                if isinstance(start, str):
                    start_date = parser.parse(start)
                elif isinstance(start, datetime):
                    start_date = start
                else:
                    raise ValueError(
                        f"Start argument ({start}) is not a valid datetime ({type(start)})"
                    )
                start_date = start_date.replace(tzinfo=None, hour=0, minute=0, second=0)
            # otherwise use default data
            else:
                start_date = self.start_datetime.replace(tzinfo=None)
            if end is not None:
                if isinstance(end, str):
                    end_date = parser.parse(end)
                elif isinstance(end, datetime):
                    end_date = end
                else:
                    raise ValueError(
                        f"End argument ({end}) is not a valid datetime ({type(end)})"
                    )
                end_date = end_date.replace(tzinfo=None)
            else:
                end_date = datetime.utcnow()

            # generate range
            date_range = create_intervals(
                start_date, end_date, interval=7, round_days=True
            )

            # retrieve data accordingly
            for dates in date_range:
                cycle_params = {
                    "start": whoop_time_str(dates[0]),
                    "end": whoop_time_str(dates[1]),
                }
                json_data = self.pull_api(
                    self._create_url("cycles"), params=cycle_params
                )

                results.append(json_data)
        else:
            raise RuntimeError("Please run the authorization function first")

        # return all data
        return results

    def get_keydata(self, raw_data=None, start=None, end=None):
        """
        This function returns a dataframe of WHOOP metrics for each day of WHOOP membership.
        In the resulting dataframe, each day is a row and contains strain, recovery, and sleep information
        """
        # retrieve all raw data and convert to dataframes
        if raw_data is None:
            raw_data = self.get_keydata_raw(start, end)
        df_data = [pd.json_normalize(data) for data in raw_data]
        all_data = pd.concat(df_data)

        # check length
        if len(all_data) == 0:
            return None

        # update the dataframe
        all_data.reset_index(drop=True, inplace=True)

        # fixing the day column so it's not a list
        all_data["days"] = all_data["days"].map(lambda d: d[0])
        all_data.rename(columns={"days": r"day"}, inplace=True)

        # Putting all time into minutes instead of milliseconds
        sleep_cols = [
            "qualityDuration",
            "needBreakdown.baseline",
            "needBreakdown.debt",
            "needBreakdown.naps",
            "needBreakdown.strain",
            "needBreakdown.total",
        ]
        for sleep_col in sleep_cols:
            if sleep_col in all_data.columns:
                all_data["sleep." + sleep_col] = (
                    all_data["sleep." + sleep_col]
                    .astype(float)
                    .apply(lambda x: np.nan if np.isnan(x) else x / 60000)
                )
            else:
                all_data["sleep." + sleep_col] = np.nan

        # Making nap variable
        all_data["nap_duration"] = all_data["sleep.naps"].apply(
            lambda x: x[0]["qualityDuration"] / 60000
            if len(x) == 1
            else (
                sum(
                    [
                        y["qualityDuration"]
                        for y in x
                        if y["qualityDuration"] is not None
                    ]
                )
                / 60000
                if len(x) > 1
                else 0
            )
        )
        all_data.drop(["sleep.naps"], axis=1, inplace=True)

        # dropping duplicates subsetting because of list columns
        all_data.drop_duplicates(subset=["day", "sleep.id"], inplace=True)

        return all_data

    def get_sports(self):
        """Retrieve a list of all sports"""
        sports = self.pull_api(self._create_url("sports", user=False))
        sport_dict = {sport["id"]: sport["name"] for sport in sports}
        self.sport_dict = self.sport_dict
        return sport_dict

    def _apply_zone(self, item: Dict[str, float], zone: int):
        if item is None or zone not in item:
            return None
        return item[zone] / 60000.0

    def get_activities(
        self, all_data=None, update_sport_dict=False, start=None, end=None
    ):
        """
        Activity data is pulled through the get_keydata functions so if the data pull is present, this function
        just transforms the activity column into a dataframe of activities, where each activity is a row.
        If it has not been pulled, this function runs the key data function then returns the activity dataframe
        """
        # pull data from all activities that have been logged so far
        if self.sport_dict and update_sport_dict is False:
            sport_dict = self.sport_dict
        else:
            sport_dict = self.get_sports()

        # make sure user is logged in
        if self.start_datetime:
            # check if existing data is used
            if all_data is not None:
                data = all_data
            else:
                data = self.get_keydata(start=start, end=end)

            # ensure data has the right format
            if data["strain.workouts"].dtype == "object":
                data["strain.workouts"] = data["strain.workouts"].apply(
                    lambda x: eval(str(x))
                )

            # retrieve the data
            data = data[data["strain.workouts"].apply(len) > 0]
            data = data.explode("strain.workouts")
            # data["strain.workouts"] = pd.json_normalize(data["strain.workouts"])
            # act_data = data["strain.workouts"].apply(pd.Series)
            act_data = pd.json_normalize(data["strain.workouts"])
            # act_data["day"] = data["day"]

            # check if there is data
            if len(data) == 0:
                return None

            # update the data
            act_data[["during.upper", "during.lower"]] = act_data[
                ["during.upper", "during.lower"]
            ].apply(pd.to_datetime)
            act_data["total_minutes"] = act_data.apply(
                lambda x: (x["during.upper"] - x["during.lower"]).total_seconds()
                / 60.0,
                axis=1,
            )
            for z in range(0, 6):
                act_data["zone{}_minutes".format(z + 1)] = act_data["zones"].apply(
                    lambda x: self._apply_zone(x, z)
                )
            act_data["sport_name"] = act_data.sportId.apply(lambda x: sport_dict[x])

            act_data["day"] = act_data["during.lower"].dt.strftime(DATE_FORMAT)
            act_data.drop(["zones", "during.bounds"], axis=1, inplace=True)
            act_data.drop_duplicates(inplace=True)
            self.all_activities = act_data
            return act_data
        else:
            raise RuntimeError("Please run the authorization function first")

    def get_sleep(self, all_data=None, start=None, end=None):
        """
        This function returns all sleep metrics in a data frame, for the duration of user's WHOOP membership.
        Each row in the data frame represents one night of sleep
        """
        if self.auth_token:
            # check if previous data can be used
            if all_data is not None:
                data = all_data
            else:
                data = self.get_keydata(start=None, end=None)

            # getting all the sleep ids
            sleep_ids = data["sleep.id"].values.tolist()
            sleep_list = [int(x) for x in sleep_ids if pd.isna(x) is False]
            all_sleep = pd.DataFrame()
            for s in sleep_list:
                m = self.pull_sleep_main(s)
                all_sleep = pd.concat([all_sleep, m])

            if len(all_sleep) == 0:
                return None

            # Cleaning sleep data
            sleep_update = [
                "qualityDuration",
                "latency",
                "debtPre",
                "debtPost",
                "needFromStrain",
                "sleepNeed",
                "habitualSleepNeed",
                "timeInBed",
                "lightSleepDuration",
                "slowWaveSleepDuration",
                "remSleepDuration",
                "wakeDuration",
                "arousalTime",
                "noDataDuration",
                "creditFromNaps",
                "projectedSleep",
            ]

            for col in sleep_update:
                if col in all_sleep.columns:
                    all_sleep[col] = (
                        all_sleep[col]
                        .astype(float)
                        .apply(lambda x: np.nan if np.isnan(x) else x / 60000)
                    )
                else:
                    all_data[col] = np.nan

            for col in ["during.bounds", "events"]:
                if col in all_sleep:
                    all_sleep.drop([col], axis=1, inplace=True)
            return all_sleep
        else:
            raise RuntimeError("Please run the authorization function first")

    def get_sleep_events_all(self, all_data=None, all_sleep=None, start=None, end=None):
        """
        This function returns all sleep events in a data frame, for the duration of user's WHOOP membership.
        Each row in the data frame represents an individual sleep event within an individual night of sleep.
        Sleep events can be joined against the sleep or main datasets by sleep id.
        All sleep times are returned in minutes.
        """
        if self.auth_token:
            # check for previous data
            if all_data is not None:
                data = all_data
            else:
                data = self.get_keydata(start=None, end=None)

            # getting all the sleep ids
            if all_sleep is not None:
                sleep_events = all_sleep[["activityId", "events"]]
                all_sleep_events = pd.concat(
                    [
                        pd.concat(
                            [
                                pd.json_normalize(events),
                                pd.DataFrame({"id": len(events) * [sleep]}),
                            ],
                            axis=1,
                        )
                        for events, sleep in zip(
                            sleep_events["events"], sleep_events["activityId"]
                        )
                    ]
                )
            else:
                sleep_ids = data["sleep.id"].values.tolist()
                sleep_list = [int(x) for x in sleep_ids if pd.isna(x) == False]
                all_sleep_events = pd.DataFrame()
                for s in sleep_list:
                    events = self.pull_sleep_events(s)
                    all_sleep_events = pd.concat([all_sleep_events, events])

            # Cleaning sleep events data
            all_sleep_events["during.lower"] = pd.to_datetime(
                all_sleep_events["during.lower"]
            )
            all_sleep_events["during.upper"] = pd.to_datetime(
                all_sleep_events["during.upper"]
            )
            all_sleep_events.drop(["during.bounds"], axis=1, inplace=True)
            all_sleep_events["total_minutes"] = all_sleep_events.apply(
                lambda x: (x["during.upper"] - x["during.lower"]).total_seconds()
                / 60.0,
                axis=1,
            )

            return all_sleep_events
        else:
            raise RuntimeError("Please run the authorization function first")

    def get_hr(self, df=False, start=None, end=None):
        """
        This function will pull every heart rate measurement recorded for the life of WHOOP membership.
        The default return for this function is a list of lists, where each "row" contains the date, time, and hr value.
        The measurements are spaced out every ~6 seconds on average.

        To return a dataframe, set df=True. This will take a bit longer, but will return a data frame.

        NOTE: This api pull takes about 6 seconds per week of data ... or 1 minutes for 10 weeks of data,
        so be careful when you pull, it may take a while.
        """
        if self.start_datetime:
            # generate date range
            date_range = create_intervals(start, end, interval=6, round_days=True)

            # create request for date range
            hr_list = []
            for dates in date_range:
                params = {
                    "start": whoop_time_str(dates[0]),
                    "end": whoop_time_str(dates[1]),
                    "order": "t",
                    "step": 6,
                }
                try:
                    hr_vals = self.pull_api(
                        self._create_url("metrics/heart_rate"), params=params
                    )["values"]
                except IOError:
                    print(f"Unable to pull data from {dates[0]} to {dates[1]}")
                    logging.warning(
                        f"Unable to pull data from {dates[0]} to {dates[1]}"
                    )
                    continue
                hr_values = [
                    [
                        datetime.utcfromtimestamp(h["time"] / 1e3).date(),
                        datetime.utcfromtimestamp(h["time"] / 1e3).time(),
                        h["data"],
                    ]
                    for h in hr_vals
                ]
                hr_list.extend(hr_values)

            # check length
            if len(hr_list) == 0:
                return None

            # check conversion
            if df:
                hr_df = pd.DataFrame(hr_list)
                hr_df.columns = ["date", "time", "hr"]
                hr_df = hr_df.reset_index()[["date", "time", "hr"]]
                return hr_df
            else:
                return hr_list
        else:
            raise RuntimeError("Please run the authorization function first")



"""Builds the various handlers for the Whoop API.

Copyright (c) 2022 Felix Geilert
"""

from typing import Any, Dict, List, Tuple, Type
import requests

import pandas as pd
import time_helper as th

from ..models import models_v1 as models


class WhoopHandler:
    def __init__(self, client) -> None:
        self.client = client

    def _check_datetime(self, date: str) -> str:
        """Checks if the given date is in the correct format."""
        new_date = th.any_to_datetime(date)

        if date is None:
            raise ValueError(f"Invalid Date provided: {date}")

        # make sure this is converted to correct format used by whoop
        return new_date.isoformat() + "Z"

    def _get(self, path: str, params: dict = None, **kwargs) -> requests.Response:
        """Sends a GET request to the Whoop API."""
        path = f"{self.client._base_path}/{path}"
        return self.client.session.get(path, params=params, **kwargs)

    def _post(self, path: str, data: dict = None, **kwargs) -> requests.Response:
        """Sends a POST request to the Whoop API."""
        path = f"{self.client._base_path}/{path}"
        return self.client.session.post(path, data=data, **kwargs)

    def _verify(self, res: requests.Response) -> Dict[str, Any]:
        """Verifies the response from the Whoop API."""
        if res.status_code != 200:
            raise Exception(f"Whoop API returned status code {res.status_code}.")
        return res.json()


class WhoopUserHandler(WhoopHandler):
    def __init__(self, client) -> None:
        super().__init__(client)

    def profile(self) -> models.UserProfile:
        res = self._get("user/profile/basic")
        data = self._verify(res)

        return models.UserProfile(**data)

    def body_measurements(self) -> models.UserMeasurements:
        res = self._get("user/body_measurements")
        data = self._verify(res)

        return models.UserMeasurements(**data)


class WhoopDataHandler(WhoopHandler):
    """Abstract class that works as a data handler

    Note: '@' symbol is used in the paths to indicate split point for id
    """

    def __init__(
        self,
        client: Any,
        path: str,
        model: Type[models.UserData],
        path_single: str = None,
    ) -> None:
        super().__init__(client)
        self._path = path
        self._path_single = path_single or path + "/@"
        self._model = model

    def _get_data(
        self,
        path: str,
        start: str = None,
        end: str = None,
        next: str = None,
        limit: int = 10,
    ) -> Tuple[Dict, str]:
        """Gets the data from the Whoop API."""
        params = self._params(start, end, next, limit)
        res = self._get(path, params=params)
        data = self._verify(res)
        return data["records"], data.get("next_token")

    def _to_df(self, data: List[models.UserData]) -> pd.DataFrame:
        """Converts the given data to a pandas DataFrame."""
        return pd.json_normalize([d.dict() for d in data])

    def _params(
        self, start: str = None, end: str = None, next: str = None, limit: int = 10
    ) -> Dict[str, Any]:
        # check limit value
        if limit > 25 or limit < 1:
            raise ValueError("Limit must be between 1 and 25.")

        # generate params
        params = {"limit": limit}
        if start:
            params["start"] = self._check_datetime(start)
        if end:
            params["end"] = self._check_datetime(end)
        if next:
            params["nextToken"] = next

        return params

    def single(self, id: int) -> models.UserData:
        """Gets a single data object from the Whoop API."""
        path = self._path_single.split("@", 1)
        path = f"{path[0]}{id}{path[1] if len(path) > 1 else ''}"
        res = self._get(path)
        data = self._verify(res)
        return self._model.from_dict(data)

    def collection(
        self,
        start: str = None,
        end: str = None,
        next: str = None,
        limit: int = 10,
        get_all_pages: bool = True,
    ) -> Tuple[List[models.UserData], str]:
        """Gets a collection of data from the Whoop API."""
        recs, token = self._get_data(self._path, start, end, next, limit)
        items = [self._model.from_dict(c) for c in recs]

        # get more data if there is a next token
        if get_all_pages:
            while token:
                data, token = self.collection(
                    start=start, end=end, next=token, limit=limit, get_all_pages=False
                )
                items.extend(data)

        return items, token

    def collection_df(
        self,
        start: str = None,
        end: str = None,
        next: str = None,
        limit: int = 10,
        get_all_pages: bool = True,
    ) -> Tuple[pd.DataFrame, str]:
        """Gets a collection of data from the Whoop API."""
        recs, token = self.collection(start, end, next, limit, get_all_pages)
        df = self._to_df(recs)

        return df, token

    def latest(self) -> models.UserData:
        """Gets the latest data from the Whoop API."""
        recs, _ = self.collection(limit=1, get_all_pages=False)
        return recs[0]


class WhoopCycleHandler(WhoopDataHandler):
    def __init__(self, client) -> None:
        super().__init__(client, "cycle", models.UserCycle)


class WhoopSleepHandler(WhoopDataHandler):
    def __init__(self, client) -> None:
        super().__init__(client, "activity/sleep", models.UserSleep)


class WhoopRecoveryHandler(WhoopDataHandler):
    def __init__(self, client) -> None:
        super().__init__(client, "recovery", models.UserRecovery, "cycle/@/recovery")


class WhoopWorkoutHandler(WhoopDataHandler):
    def __init__(self, client) -> None:
        super().__init__(client, "activity/workout", models.UserWorkout)



import os
from setuptools import setup
from setuptools import find_packages
from whoopy import __version__
import pathlib


__status__ = "Package"
__copyright__ = "Copyright 2022"
__license__ = "MIT License"
__author__ = "Felix Geilert"


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "readme.md"), encoding="utf-8") as f:
    long_description = f.read()


def versions_in_requirements(file):
    lines = file.read().splitlines()
    versions = [line for line in lines if not line.isspace() and "--" not in line]
    return list(versions)


HERE = pathlib.Path(__file__).parent
with open(HERE / "requirements.txt") as f:
    required_list = versions_in_requirements(f)

setup(
    name="whoopy",
    version=__version__,
    description="Unofficial Client for the Whoop API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="whoop api",
    url="https://github.com/felixnext/whoopy",
    download_url=f"https://github.com/felixnext/whoopy/releases/tag/v{__version__}",
    author="Felix Geilert",
    license="MIT License",
    packages=find_packages(include=["whoopy", "whoopy.*"]),
    install_requires=required_list,
    setup_requires=["pytest-runner", "flake8"],
    tests_require=["pytest"],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



"""Official Whoop API.

Generated against version 1.0 of the API.

Copyright 2022 (C) Felix Geilert
"""

from typing import Tuple, List
from typing_extensions import Self
import webbrowser
import uuid
import requests
import json
import os

from .handlers import handler_v1 as handlers


API_VERSION = "1"
API_BASE = "https://api.prod.whoop.com/"
API_AUTH = f"{API_BASE}oauth/oauth2"
SCOPES = [
    "offline",
    "read:recovery",
    "read:cycles",
    "read:sleep",
    "read:workout",
    "read:profile",
    "read:body_measurement",
]


class WhoopClient:
    def __init__(
        self,
        access_token: str,
        expires_in: int,
        scopes: List[str],
        refresh_token: str = None,
        client_id: str = None,
        client_secret: str = None,
    ):
        """Creates a new WhoopClient.

        Note that when client_id and client_secret are not provided, the refresh function will not work.

        Args:
            access_token (str): The access token.
            expires_in (int): The time until the token expires (in seconds).
            scopes (List[str]): The scopes that the token has.
            refresh_token (str, optional): The refresh token. Defaults to None.
            client_id (str, optional): The client ID. Defaults to None.
            client_secret (str, optional): The client secret. Defaults to None.
        """
        self.token = access_token
        self.expires_in = expires_in
        self.scopes = scopes
        self.refresh_token = refresh_token
        self._client_id = client_id
        self._client_secret = client_secret

        # create a session
        self.user_agent = "Python/3.X (X11; Linux x86_64)"
        self._update_session()

        # create a bunch of handlers
        self.user = handlers.WhoopUserHandler(self)
        self.cycle = handlers.WhoopCycleHandler(self)
        self.sleep = handlers.WhoopSleepHandler(self)
        self.workout = handlers.WhoopWorkoutHandler(self)
        self.recovery = handlers.WhoopRecoveryHandler(self)

    @property
    def _token(self):
        return {
            "access_token": self.token,
            "expires_in": self.expires_in,
            "refresh_token": self.refresh_token,
            "scopes": self.scopes,
        }

    def _update_session(self):
        """Updates the session with the new token."""
        self._base_path = f"{API_BASE}developer/v1"
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Authorization": f"Bearer {self.token}",
                "User-Agent": self.user_agent,
            }
        )

    def store_token(self, path: str):
        """Stores the token to a file.

        Args:
            path (str): The path to the file (e.g. ".tokens/token.json").
        """
        # verify that folder exists
        base_dir = os.path.dirname(path)
        os.makedirs(base_dir, exist_ok=True)

        # store the token
        with open(path, "w") as f:
            json.dump(self._token, f)

    @classmethod
    def from_token(cls, path: str, client_id: str, client_secret: str) -> Self:
        """Loads a token from a file.

        Args:
            path (str): The path to the file (e.g. ".tokens/token.json").
            client_id (str): The client ID.
            client_secret (str): The client secret.
        """
        with open(path, "r") as f:
            token = json.load(f)
        client = cls(
            token["access_token"],
            token["expires_in"],
            token["scopes"],
            token["refresh_token"],
            client_id,
            client_secret,
        )
        client.refresh()

        return client

    # retrieves the authorization url
    @classmethod
    def auth_url(
        cls,
        client_id: str,
        client_secret: str,
        redirect_uri: str,
        state: str = None,
        scopes: List[str] = None,
    ) -> Tuple[str, str]:
        """Generates authorization url for the Whoop API."""
        # check state
        if not state:
            state = str(uuid.uuid4())

        # retrieve the data
        if len(state) < 8:
            raise ValueError("State must be at least 8 characters long")

        # generate the
        scopes = scopes or SCOPES
        scope = " ".join(scopes)
        res = (
            f"{API_AUTH}/auth?"
            f"scope={scope}&"
            f"client_id={client_id}&"
            f"client_secret={client_secret}&"
            f"state={state}&"
            f"redirect_uri={redirect_uri}&"
            "response_type=code"
        )

        return res, state

    @classmethod
    def _parse_token(cls, payload, scopes):
        url = f"{API_AUTH}/token"

        # retrieve the codes
        res = requests.post(url, data=payload)
        if res.status_code != 200:
            raise RuntimeError(f"Authorization failed with code {res.status_code}")
        codes = res.json()
        token_scopes = codes["scope"].split(" ")

        # validate scopes
        if scopes:
            for scope in scopes:
                if scope not in token_scopes:
                    raise ValueError(f"Scope {scope} not granted")

        return codes, token_scopes

    @classmethod
    def authorize(
        cls,
        code: str,
        client_id: str,
        client_secret: str,
        redirect_url: str = "https://jwt.ms/",
        scopes: List[str] = None,
    ) -> Self:
        """Authorize the client with the given code."""
        # generate request using the code
        payload = {
            "client_id": client_id,
            "client_secret": client_secret,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirect_url,
        }
        codes, token_scopes = cls._parse_token(payload, scopes)

        # generate the client
        return cls(
            codes["access_token"],
            codes["expires_in"],
            token_scopes,
            codes.get("refresh_token", None),
            client_id=client_id,
            client_secret=client_secret,
        )

    @classmethod
    def auth_flow(
        cls,
        client_id: str,
        client_secret: str,
        redirect_url: str = "https://jwt.ms/",
        state: str = None,
        scopes: List[str] = None,
    ) -> Self:
        """Runs through the entire auth flow.

        Note: This requires to copy the code query attribute from the resulting redirect url.

        Args:
            client_id (str): The client ID.
            client_secret (str): The client secret.
            redirect_url (str, optional): The redirect URL. Defaults to "https://jwt.ms/".
            state (str, optional): The state passed through to the output of request. Defaults to None.
            scopes (List[str], optional): The scopes to request. Defaults to None.
                (In case of None, all scopes are requested.)

        Returns:
            The client object.
        """
        # retrieve url
        url, state = cls.auth_url(client_id, client_secret, redirect_url, state, scopes)

        # send user to auth and wait for code
        webbrowser.open(url)
        code = input("Copy Code Attribute from URL: ")

        # authorize
        client = cls.authorize(code, client_id, client_secret, redirect_url, scopes)

        # complete
        return client

    def refresh(self):
        """Refreshes the token provided."""
        # verify client is setup correctly
        if self.refresh_token is None:
            raise ValueError("No refresh token provided")
        if self._client_id is None or self._client_secret is None:
            raise ValueError("No client id or secret provided")

        # generate request using the code
        payload = {
            "client_id": self._client_id,
            "client_secret": self._client_secret,
            "grant_type": "refresh_token",
            "refresh_token": self.refresh_token,
        }

        # retrieve the codes
        codes, _ = self._parse_token(payload, self.scopes)

        # update data
        self.token = codes["access_token"]
        self.expires_in = codes["expires_in"]
        self.refresh_token = codes.get("refresh_token", None)

        # update sess
        self._update_session()
